#include "ros/ros.h"

#include <geometry_msgs/Pose.h>
#include <std_msgs/Int16.h>


#include "iri_wam_common_msgs/QueryWamMoveInCartesian.h"	//move arm in cartesian
#include "iri_common_drivers_msgs/tool_closeAction.h"
#include "iri_common_drivers_msgs/tool_openAction.h"
#include <actionlib/client/simple_action_client.h>
#include <actionlib/client/terminal_state.h>

// #include "iri_common_drivers_msgs/QueryJointsMovement.h"

// #include "iri_wam_common_msgs/SimpleBhandPickup.h"


std::vector<double> stackPose_Vtr(7);	//Vector holding the value of the pose
std::vector<double> c1Pose_Vtr(7);
int cellSize;

bool isObjectPicked = false;
bool isGoalReceived = false;
int objType, goalCell;
int moveCount = 0;
geometry_msgs::Pose stackPose; //Position and orientation of the place where objects are stacked
geometry_msgs::Pose c1Pose;		//Position and orientation of the 1st cell of 3x3 grid for tictactoe
geometry_msgs::Pose goalPose;  

//function to convert vector (retrieve from params) to geometry_msgs::Pose
geometry_msgs::Pose vector2Pose(std::vector<double> _vtr){
	geometry_msgs::Pose pose_;
	try{
		pose_.position.x = _vtr[0];
		pose_.position.y = _vtr[1];
		pose_.position.z = _vtr[2];
		pose_.orientation.x = _vtr[3];
		pose_.orientation.y = _vtr[4];
		pose_.orientation.z = _vtr[5];
		pose_.orientation.w = _vtr[6];
	}
	catch(const std::exception& e){
		ROS_ERROR("[%s]: %s",ros::this_node::getName().c_str(),e.what());		
	}	
	return pose_;
}


//Callback to get the pose of the cell
//request should be int from 1-9
// | 1 | 2 | 3 |
// | 4 | 5 | 6 |
// | 7 | 8 | 9 |
//Assumption (need checking):
//Location of Cell 1 is known
//Going from cell 1 to cell 2 will be 10cm along the x-axis
//Going from cell 1 to cell 4 will be 10cm along the y-axis
void changeStackPose(){
	stackPose.position.x += 4*cellSize;
	stackPose.position.y += moveCount*cellSize;
}

void goalCell_CB(const std_msgs::Int16::ConstPtr& msg){
	goalCell = msg->data;

	double y_trans = (goalCell-1)/3;
	double x_trans = (goalCell - 3*y_trans);

	goalPose = c1Pose;
	c1Pose.position.x += x_trans*cellSize;
	c1Pose.position.y += y_trans*cellSize;

	changeStackPose();
	isGoalReceived = true;

}


int main(int argc, char** argv){
	ros::init(argc, argv, "ttt_executor_node");

	ros::NodeHandle nh;

	if(!nh.getParam("cell_size", cellSize)){
		cellSize = 0.1;	//default value for cell size is 10cm
	}

	if(!nh.getParam("c1_pose", c1Pose_Vtr)){
		ROS_ERROR("[%s]: %s", ros::this_node::getName().c_str(), "Location of 1st cell is not specified");
		return -1;
	}
	else{
		c1Pose = vector2Pose(c1Pose_Vtr);
		stackPose = c1Pose;
	}




	ros::Subscriber cell_sub = nh.subscribe("target_cell", 10, goalCell_CB);

	//WAM SERVICES
	//require path to the service - unknown
	ros::ServiceClient moveArm_client = nh.serviceClient<iri_wam_common_msgs::QueryWamMoveInCartesian>("/iri_wam/");
	iri_wam_common_msgs::QueryWamMoveInCartesian srv;


	//WAM ACTION
	//need real paths - unknown
	actionlib::SimpleActionClient<iri_common_drivers_msgs::tool_closeAction> closeAct("/tool_close",true);
	actionlib::SimpleActionClient<iri_common_drivers_msgs::tool_openAction> openAct("/tool_open",true);
	closeAct.waitForServer();
	openAct.waitForServer();

	iri_common_drivers_msgs::tool_closeGoal closeGoal;
	iri_common_drivers_msgs::tool_openGoal openGoal;

	ros::Rate rate(10.0);
	while(ros::ok()){
		ros::spinOnce();
		if(isGoalReceived){

			//pickup stage - go to stack place
			if(!isObjectPicked){
				srv.request.pose = stackPose;
				srv.request.vel = 0.5;
				srv.request.acc = 0.5;
				if(moveArm_client.call(srv)){
					ROS_INFO("Arm moved to requested pose");
					closeAct.sendGoal(closeGoal);
					if(closeAct.waitForResult(ros::Duration(10.0))){
						actionlib::SimpleClientGoalState state = closeAct.getState();
						ROS_INFO("Tool_close action: %s",state.toString().c_str());
						isObjectPicked = true;
					}
					else{
						ROS_WARN("Tool_close action failed.");
					}
				}
				

				else
					ROS_ERROR("Failed to move arm to requested pose");
			}

			//move to goal stage
			else{
				srv.request.pose = goalPose;
				srv.request.vel = 0.5;
				srv.request.acc = 0.5;
				if(moveArm_client.call(srv)){
					ROS_INFO("Arm moved to requested pose");
					openAct.sendGoal(openGoal);
					if(openAct.waitForResult(ros::Duration(10.0))){
						actionlib::SimpleClientGoalState state = openAct.getState();
						ROS_INFO("Tool_open action: %s",state.toString().c_str());
						isGoalReceived = false;
						isObjectPicked = false;
					}
					else{
						ROS_WARN("Tool_open action failed");
					}
				}
				else
					ROS_ERROR("Failed to move arm to requested pose");				
			}



		}
		rate.sleep();
	}
	return 0;

}